#!/usr/bin/env python
# -*- coding: utf-8 -*-

import hashlib
import logging
import os
import re
import requests

from collections import namedtuple
from os import environ, makedirs
from os.path import exists, expanduser, join
from typing import Mapping, Optional, Union

import numpy as np
import torch

from torch import nn, Tensor
from tqdm import tqdm
from transformers import logging as tfmers_logging
from transformers.models.auto.modeling_auto import AutoModel
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.tokenization_utils_fast import PreTrainedTokenizerFast

from aac_metrics.utils.checks import is_mono_sents
from aac_metrics.utils.globals import _get_device


# config according to the settings on your computer, this should be default setting of shadowsocks
DEFAULT_PROXIES = {
    "http": "socks5h://127.0.0.1:1080",
    "https": "socks5h://127.0.0.1:1080",
}
PRETRAIN_ECHECKERS_DICT = {
    "echecker_clotho_audiocaps_base": (
        "https://github.com/blmoistawinde/fense/releases/download/V0.1/echecker_clotho_audiocaps_base.ckpt",
        "1a719f090af70614bbdb9f9437530b7e133c48cfa4a58d964de0d47fc974a2fa",
    ),
    "echecker_clotho_audiocaps_tiny": (
        "https://github.com/blmoistawinde/fense/releases/download/V0.1/echecker_clotho_audiocaps_tiny.ckpt",
        "90ed0ac5033ec497ec66d4f68588053813e085671136dae312097c96c504f673",
    ),
}

RemoteFileMetadata = namedtuple("RemoteFileMetadata", ["filename", "url", "checksum"])

pylog = logging.getLogger(__name__)


ERROR_NAMES = (
    "add_tail",
    "repeat_event",
    "repeat_adv",
    "remove_conj",
    "remove_verb",
    "error",
)


class BERTFlatClassifier(nn.Module):
    def __init__(self, model_type: str, num_classes: int = 5) -> None:
        super().__init__()
        self.model_type = model_type
        self.num_classes = num_classes
        self.encoder = AutoModel.from_pretrained(model_type)
        self.dropout = nn.Dropout(self.encoder.config.hidden_dropout_prob)
        self.clf = nn.Linear(self.encoder.config.hidden_size, num_classes)

    @classmethod
    def from_pretrained(
        cls,
        model_name: str = "echecker_clotho_audiocaps_base",
        device: Union[str, torch.device, None] = "auto",
        use_proxy: bool = False,
        proxies: Optional[dict[str, str]] = None,
        verbose: int = 0,
    ) -> "BERTFlatClassifier":
        return __load_pretrain_echecker(model_name, device, use_proxy, proxies, verbose)

    def forward(
        self,
        input_ids: Optional[Tensor] = None,
        attention_mask: Optional[Tensor] = None,
        token_type_ids: Optional[Tensor] = None,
        **kwargs,
    ) -> Tensor:
        outputs = self.encoder(input_ids, attention_mask, token_type_ids)
        x = outputs.last_hidden_state[:, 0, :]
        x = self.dropout(x)
        logits = self.clf(x)
        return logits


def fer(
    candidates: list[str],
    return_all_scores: bool = True,
    echecker: Union[str, BERTFlatClassifier] = "echecker_clotho_audiocaps_base",
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    error_threshold: float = 0.9,
    device: Union[str, torch.device, None] = "auto",
    batch_size: int = 32,
    reset_state: bool = True,
    return_probs: bool = False,
    verbose: int = 0,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Tensor]]]:
    """Return Fluency Error Rate (FER) detected by a pre-trained BERT model.

    - Paper: https://arxiv.org/abs/2110.04684
    - Original implementation: https://github.com/blmoistawinde/fense

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param echecker: The echecker model used to detect fluency errors.
        Can be "echecker_clotho_audiocaps_base", "echecker_clotho_audiocaps_tiny", "none" or None.
        defaults to "echecker_clotho_audiocaps_base".
    :param echecker_tokenizer: The tokenizer of the echecker model.
        If None and echecker is not None, this value will be inferred with `echecker.model_type`.
        defaults to None.
    :param error_threshold: The threshold used to detect fluency errors for echecker model. defaults to 0.9.
    :param device: The PyTorch device used to run FENSE models. If "auto", it will use cuda if available. defaults to "auto".
    :param batch_size: The batch size of the echecker models. defaults to 32.
    :param reset_state: If True, reset the state of the PyTorch global generator after the initialization of the pre-trained models. defaults to True.
    :param return_probs: If True, return each individual error probability given by the fluency detector model. defaults to False.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    if not is_mono_sents(candidates):
        error_msg = f"Invalid candidates type. (expected list[str], found {candidates.__class__.__name__})"
        raise ValueError(error_msg)

    # Init models
    echecker, echecker_tokenizer = _load_echecker_and_tokenizer(
        echecker, echecker_tokenizer, device, reset_state, verbose
    )

    # Compute and apply fluency error detection penalty
    probs_outs_sents = __detect_error_sents(
        echecker,
        echecker_tokenizer,  # type: ignore
        candidates,
        batch_size,
        device,
    )
    fer_scores = (probs_outs_sents["error"] > error_threshold).astype(float)

    fer_scores = torch.from_numpy(fer_scores)
    fer_score = fer_scores.mean()

    if return_all_scores:
        fer_outs_corpus = {
            "fer": fer_score,
        }
        fer_outs_sents = {
            "fer": fer_scores,
        }

        if return_probs:
            probs_outs_sents = {f"fer.{k}_prob": v for k, v in probs_outs_sents.items()}
            probs_outs_sents = {
                k: torch.from_numpy(v) for k, v in probs_outs_sents.items()
            }
            probs_outs_corpus = {k: v.mean() for k, v in probs_outs_sents.items()}

            fer_outs_corpus = probs_outs_corpus | fer_outs_corpus
            fer_outs_sents = probs_outs_sents | fer_outs_sents

        fer_outs = fer_outs_corpus, fer_outs_sents

        return fer_outs
    else:
        return fer_score


# - Private functions
def _load_echecker_and_tokenizer(
    echecker: Union[str, BERTFlatClassifier] = "echecker_clotho_audiocaps_base",
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    device: Union[str, torch.device, None] = "auto",
    reset_state: bool = True,
    verbose: int = 0,
) -> tuple[BERTFlatClassifier, AutoTokenizer]:
    state = torch.random.get_rng_state()

    device = _get_device(device)
    if isinstance(echecker, str):
        echecker = __load_pretrain_echecker(
            echecker_model=echecker, device=device, verbose=verbose
        )

    if echecker_tokenizer is None:
        echecker_tokenizer = AutoTokenizer.from_pretrained(echecker.model_type)  # type: ignore

    echecker = echecker.eval()
    for p in echecker.parameters():
        p.requires_grad_(False)

    if reset_state:
        torch.random.set_rng_state(state)

    return echecker, echecker_tokenizer  # type: ignore


def __detect_error_sents(
    echecker: BERTFlatClassifier,
    echecker_tokenizer: PreTrainedTokenizerFast,
    sents: list[str],
    batch_size: int,
    device: Union[str, torch.device, None],
    max_len: int = 64,
) -> dict[str, np.ndarray]:
    device = _get_device(device)

    if len(sents) <= batch_size:
        batch = __infer_preprocess(
            echecker_tokenizer,
            sents,
            max_len=max_len,
            device=device,
            dtype=torch.long,
        )
        logits: Tensor = echecker(**batch)
        assert not logits.requires_grad
        # batch_logits: (bsize, num_classes=6)
        # note: fix error in the original fense code: https://github.com/blmoistawinde/fense/blob/main/fense/evaluator.py#L69
        probs = logits.sigmoid().transpose(0, 1).cpu().numpy()
        probs_dic: dict[str, np.ndarray] = dict(zip(ERROR_NAMES, probs))

    else:
        dic_lst_probs = {name: [] for name in ERROR_NAMES}

        for i in range(0, len(sents), batch_size):
            batch = __infer_preprocess(
                echecker_tokenizer,
                sents[i : i + batch_size],
                max_len=max_len,
                device=device,
                dtype=torch.long,
            )

            batch_logits: Tensor = echecker(**batch)
            assert not batch_logits.requires_grad
            # batch_logits: (bsize, num_classes=6)
            # classes: add_tail, repeat_event, repeat_adv, remove_conj, remove_verb, error
            probs = batch_logits.sigmoid().cpu().numpy()

            for j, name in enumerate(dic_lst_probs.keys()):
                dic_lst_probs[name].append(probs[:, j])

        probs_dic = {
            name: np.concatenate(probs) for name, probs in dic_lst_probs.items()
        }

    return probs_dic


def __check_download_resource(
    remote: RemoteFileMetadata,
    use_proxy: bool = False,
    proxies: Optional[dict[str, str]] = None,
) -> str:
    proxies = DEFAULT_PROXIES if use_proxy and proxies is None else proxies
    data_home = __get_data_home()
    file_path = os.path.join(data_home, remote.filename)
    if not os.path.exists(file_path):
        # currently don't capture error at this level, assume download success
        file_path = __download(remote, data_home, use_proxy, proxies)
    return file_path


def __infer_preprocess(
    tokenizer: PreTrainedTokenizerFast,
    texts: list[str],
    max_len: int,
    device: Union[str, torch.device, None],
    dtype: torch.dtype,
) -> Mapping[str, Tensor]:
    device = _get_device(device)
    texts = __text_preprocess(texts)  # type: ignore
    batch = tokenizer(texts, truncation=True, padding="max_length", max_length=max_len)
    for k in ("input_ids", "attention_mask", "token_type_ids"):
        batch[k] = torch.as_tensor(batch[k], device=device, dtype=dtype)  # type: ignore
    return batch


def __download(
    remote: RemoteFileMetadata,
    file_path: Optional[str] = None,
    use_proxy: bool = False,
    proxies: Optional[dict[str, str]] = DEFAULT_PROXIES,
) -> str:
    data_home = __get_data_home()
    file_path = __fetch_remote(remote, data_home, use_proxy, proxies)
    return file_path


def __download_with_bar(
    url: str,
    file_path: str,
    proxies: Optional[dict[str, str]] = DEFAULT_PROXIES,
) -> str:
    # Streaming, so we can iterate over the response.
    response = requests.get(url, stream=True, proxies=proxies)
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 KB
    progress_bar = tqdm(total=total_size_in_bytes, unit="B", unit_scale=True)
    with open(file_path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        raise Exception("ERROR, something went wrong with the downloading")
    return file_path


def __fetch_remote(
    remote: RemoteFileMetadata,
    dirname: Optional[str] = None,
    use_proxy: bool = False,
    proxies: Optional[dict[str, str]] = DEFAULT_PROXIES,
) -> str:
    """Helper function to download a remote dataset into path
    Fetch a dataset pointed by remote's url, save into path using remote's
    filename and ensure its integrity based on the SHA256 Checksum of the
    downloaded file.
    Parameters
    ----------
    remote : RemoteFileMetadata
        Named tuple containing remote dataset meta information: url, filename
        and checksum
    dirname : string
        Directory to save the file to.
    Returns
    -------
    file_path: string
        Full path of the created file.
    """

    file_path = remote.filename if dirname is None else join(dirname, remote.filename)
    proxies = None if not use_proxy else proxies
    file_path = __download_with_bar(remote.url, file_path, proxies)
    checksum = __sha256(file_path)
    if remote.checksum != checksum:
        raise IOError(
            "{} has an SHA256 checksum ({}) "
            "differing from expected ({}), "
            "file may be corrupted.".format(file_path, checksum, remote.checksum)
        )
    return file_path


def __get_data_home(data_home: Optional[str] = None) -> str:  # type: ignore
    """Return the path of the scikit-learn data dir.
    This folder is used by some large dataset loaders to avoid downloading the
    data several times.
    By default the data dir is set to a folder named 'fense_data' in the
    user home folder.
    Alternatively, it can be set by the 'FENSE_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.
    Parameters
    ----------
    data_home : str | None
        The path to data dir.
    """
    if data_home is None:
        data_home = environ.get("FENSE_DATA", join(torch.hub.get_dir(), "fense_data"))

    data_home: str
    data_home = expanduser(data_home)
    if not exists(data_home):
        makedirs(data_home)
    return data_home


def __load_pretrain_echecker(
    echecker_model: str,
    device: Union[str, torch.device, None] = "auto",
    use_proxy: bool = False,
    proxies: Optional[dict[str, str]] = None,
    verbose: int = 0,
) -> BERTFlatClassifier:
    if echecker_model not in PRETRAIN_ECHECKERS_DICT:
        raise ValueError(
            f"Invalid argument {echecker_model=}. (expected one of {tuple(PRETRAIN_ECHECKERS_DICT.keys())})"
        )

    device = _get_device(device)
    tfmers_logging.set_verbosity_error()  # suppress loading warnings
    url, checksum = PRETRAIN_ECHECKERS_DICT[echecker_model]
    remote = RemoteFileMetadata(
        filename=f"{echecker_model}.ckpt", url=url, checksum=checksum
    )
    file_path = __check_download_resource(remote, use_proxy, proxies)

    if verbose >= 2:
        pylog.debug(f"Loading echecker model from '{file_path}'.")

    model_states = torch.load(file_path)

    if verbose >= 2:
        pylog.debug(
            f"Loading echecker model type '{model_states['model_type']}' with '{model_states['num_classes']}' classes."
        )

    echecker = BERTFlatClassifier(
        model_type=model_states["model_type"],
        num_classes=model_states["num_classes"],
    )
    echecker.load_state_dict(model_states["state_dict"])
    echecker.eval()
    echecker.to(device=device)
    return echecker


def __sha256(path: str) -> str:
    """Calculate the sha256 hash of the file at path."""
    sha256hash = hashlib.sha256()
    chunk_size = 8192
    with open(path, "rb") as f:
        while True:
            buffer = f.read(chunk_size)
            if not buffer:
                break
            sha256hash.update(buffer)
    return sha256hash.hexdigest()


def __text_preprocess(inp: Union[str, list[str]]) -> Union[str, list[str]]:
    if isinstance(inp, str):
        return re.sub(r"[^\w\s]", "", inp).lower()
    else:
        return [re.sub(r"[^\w\s]", "", x).lower() for x in inp]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys

from typing import Optional


_TRUE_VALUES = ("true", "1", "t", "yes", "y")
_FALSE_VALUES = ("false", "0", "f", "no", "n")


def _str_to_bool(s: str) -> bool:
    s = str(s).strip().lower()
    if s in _TRUE_VALUES:
        return True
    elif s in _FALSE_VALUES:
        return False
    else:
        raise ValueError(
            f"Invalid argument {s=}. (expected one of {_TRUE_VALUES + _FALSE_VALUES})"
        )


def _str_to_opt_str(s: str) -> Optional[str]:
    s = str(s)
    if s.lower() == "none":
        return None
    else:
        return s


def _setup_logging(pkg_name: str, verbose: int, set_format: bool = True) -> None:
    handler = logging.StreamHandler(sys.stdout)
    if set_format:
        format_ = "[%(asctime)s][%(name)s][%(levelname)s] - %(message)s"
        handler.setFormatter(logging.Formatter(format_))

    pkg_logger = logging.getLogger(pkg_name)

    found = False
    for handler in pkg_logger.handlers:
        if isinstance(handler, logging.StreamHandler) and handler.stream is sys.stdout:
            found = True
            break
    if not found:
        pkg_logger.addHandler(handler)

    if verbose < 0:
        level = logging.ERROR
    elif verbose == 0:
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    else:
        level = logging.DEBUG

    pkg_logger.setLevel(level)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from unittest import TestCase

import torch

from aac_metrics import evaluate
from aac_metrics.functional import cider_d
from aac_metrics.utils.tokenization import (
    preprocess_mono_sents,
    preprocess_mult_sents,
)


class TestReadmeExamples(TestCase):
    def test_example_1(self) -> None:
        candidates: list[str] = ["a man is speaking", "rain falls"]
        mult_references: list[list[str]] = [
            [
                "a man speaks.",
                "someone speaks.",
                "a man is speaking while a bird is chirping in the background",
            ],
            ["rain is falling hard on a surface"],
        ]

        corpus_scores, _ = evaluate(candidates, mult_references)
        # print(corpus_scores)
        # dict containing the score of each metric: "bleu_1", "bleu_2", "bleu_3", "bleu_4", "rouge_l", "meteor", "cider_d", "spice", "spider"
        # {"bleu_1": tensor(0.4278), "bleu_2": ..., ...}

        expected_keys = [
            "bleu_1",
            "bleu_2",
            "bleu_3",
            "bleu_4",
            "rouge_l",
            "meteor",
            "cider_d",
            "spice",
            "spider",
        ]
        self.assertSetEqual(set(corpus_scores.keys()), set(expected_keys))
        self.assertTrue(
            torch.allclose(
                corpus_scores["bleu_1"],
                torch.as_tensor(0.4278, dtype=torch.float64),
                atol=0.0001,
            ),
            f"{corpus_scores['bleu_1']=}",
        )

    def test_example_2(self) -> None:
        candidates: list[str] = ["a man is speaking", "rain falls"]
        mult_references: list[list[str]] = [
            [
                "a man speaks.",
                "someone speaks.",
                "a man is speaking while a bird is chirping in the background",
            ],
            ["rain is falling hard on a surface"],
        ]

        corpus_scores, _ = evaluate(candidates, mult_references, metrics="dcase2023")
        # print(corpus_scores)
        # dict containing the score of each metric: "meteor", "cider_d", "spice", "spider", "spider_fl", "fer"

        expected_keys = ["meteor", "cider_d", "spice", "spider", "spider_fl", "fer"]
        self.assertTrue(set(corpus_scores.keys()).issuperset(expected_keys))

    def test_example_3(self) -> None:
        candidates: list[str] = ["a man is speaking", "rain falls"]
        mult_references: list[list[str]] = [
            [
                "a man speaks.",
                "someone speaks.",
                "a man is speaking while a bird is chirping in the background",
            ],
            ["rain is falling hard on a surface"],
        ]

        candidates = preprocess_mono_sents(candidates)
        mult_references = preprocess_mult_sents(mult_references)

        outputs: tuple[dict, dict] = cider_d(candidates, mult_references)  # type: ignore
        corpus_scores, sents_scores = outputs
        # print(corpus_scores)
        # {"cider_d": tensor(0.9614)}
        # print(sents_scores)
        # {"cider_d": tensor([1.3641, 0.5587])}

        self.assertTrue(set(corpus_scores.keys()).issuperset({"cider_d"}))
        self.assertTrue(set(sents_scores.keys()).issuperset({"cider_d"}))

        dtype = torch.float64

        self.assertTrue(
            torch.allclose(
                corpus_scores["cider_d"],
                torch.as_tensor(0.9614, dtype=dtype),
                atol=0.0001,
            )
        )
        self.assertTrue(
            torch.allclose(
                sents_scores["cider_d"],
                torch.as_tensor([1.3641, 0.5587], dtype=dtype),
                atol=0.0001,
            )
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Union

import torch

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.fer import (
    fer,
    _load_echecker_and_tokenizer,
    ERROR_NAMES,
)


pylog = logging.getLogger(__name__)


class FER(AACMetric[Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]]):
    """Return Fluency Error Rate (FER) detected by a pre-trained BERT model.

    - Paper: https://arxiv.org/abs/2110.04684
    - Original implementation: https://github.com/blmoistawinde/fense

    For more information, see :func:`~aac_metrics.functional.fer.fer`.
    """

    full_state_update = False
    higher_is_better = False
    is_differentiable = False

    min_value = -1.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        echecker: str = "echecker_clotho_audiocaps_base",
        error_threshold: float = 0.9,
        device: Union[str, torch.device, None] = "auto",
        batch_size: int = 32,
        reset_state: bool = True,
        return_probs: bool = False,
        verbose: int = 0,
    ) -> None:
        echecker, echecker_tokenizer = _load_echecker_and_tokenizer(echecker, None, device, reset_state, verbose)  # type: ignore

        super().__init__()
        self._return_all_scores = return_all_scores
        self._echecker = echecker
        self._echecker_tokenizer = echecker_tokenizer
        self._error_threshold = error_threshold
        self._device = device
        self._batch_size = batch_size
        self._reset_state = reset_state
        self._return_probs = return_probs
        self._verbose = verbose

        self._candidates = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return fer(
            candidates=self._candidates,
            return_all_scores=self._return_all_scores,
            echecker=self._echecker,
            echecker_tokenizer=self._echecker_tokenizer,
            error_threshold=self._error_threshold,
            device=self._device,
            batch_size=self._batch_size,
            reset_state=self._reset_state,
            return_probs=self._return_probs,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        hparams = {"device": self._device, "batch_size": self._batch_size}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return ("fer",) + tuple(f"fer.{name}_prob" for name in ERROR_NAMES)

    def reset(self) -> None:
        self._candidates = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        *args,
        **kwargs,
    ) -> None:
        self._candidates += candidates



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
import tempfile

from pathlib import Path
from typing import Any, Optional, Union

import torch


pylog = logging.getLogger(__name__)


# Public functions
def get_default_cache_path() -> str:
    """Returns the default cache directory path.

    If :func:`~aac_metrics.utils.globals.set_default_cache_path` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_METRICS_CACHE_PATH has been set to a string, it will return its value.
    Else it will be equal to "~/.cache" by default.
    """
    return __get_default_value("cache")


def get_default_java_path() -> str:
    """Returns the default java executable path.

    If :func:`~aac_metrics.utils.globals.set_default_java_path` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_METRICS_JAVA_PATH has been set to a string, it will return its value.
    Else it will be equal to "java" by default.
    """
    return __get_default_value("java")


def get_default_tmp_path() -> str:
    """Returns the default temporary directory path.

    If :func:`~aac_metrics.utils.globals.set_default_tmp_path` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_METRICS_TMP_PATH has been set to a string, it will return its value.
    Else it will be equal to the value returned by :func:`~tempfile.gettempdir()` by default.
    """
    return __get_default_value("tmp")


def set_default_cache_path(cache_path: Union[str, Path, None]) -> None:
    """Override default cache directory path."""
    __set_default_value("cache", cache_path)


def set_default_java_path(java_path: Union[str, Path, None]) -> None:
    """Override default java executable path."""
    __set_default_value("java", java_path)


def set_default_tmp_path(tmp_path: Union[str, Path, None]) -> None:
    """Override default temporary directory path."""
    __set_default_value("tmp", tmp_path)


# Private functions
def _get_cache_path(cache_path: Union[str, Path, None] = None) -> str:
    return __get_value("cache", cache_path)


def _get_device(
    device: Union[str, torch.device, None] = None
) -> Optional[torch.device]:
    value_name = "device"
    process_func = __DEFAULT_GLOBALS[value_name]["process"]
    device = process_func(device)
    return device  # type: ignore


def _get_java_path(java_path: Union[str, Path, None] = None) -> str:
    return __get_value("java", java_path)


def _get_tmp_path(tmp_path: Union[str, Path, None] = None) -> str:
    return __get_value("tmp", tmp_path)


def __get_default_value(value_name: str) -> Any:
    values = __DEFAULT_GLOBALS[value_name]["values"]
    process_func = __DEFAULT_GLOBALS[value_name]["process"]

    for source, value_or_env_varname in values.items():
        if source.startswith("env"):
            value = os.getenv(value_or_env_varname, None)
        else:
            value = value_or_env_varname

        if value is not None:
            value = process_func(value)
            return value

    pylog.error(f"Values: {values}")
    raise RuntimeError(
        f"Invalid default value for value_name={value_name}. (all default values are None)"
    )


def __set_default_value(
    value_name: str,
    value: Any,
) -> None:
    __DEFAULT_GLOBALS[value_name]["values"]["user"] = value


def __get_value(value_name: str, value: Any = None) -> Any:
    if value is ... or value is None:
        return __get_default_value(value_name)
    else:
        process_func = __DEFAULT_GLOBALS[value_name]["process"]
        value = process_func(value)
        return value


def __process_path(value: Union[str, Path, None]) -> Union[str, None]:
    if value is None or value is ...:
        return None
    value = str(value)
    value = osp.expanduser(value)
    value = osp.expandvars(value)
    return value


def __process_device(value: Union[str, torch.device, None]) -> Optional[torch.device]:
    if value is None or value is ...:
        return None
    if value == "auto":
        value = "cuda" if torch.cuda.is_available() else "cpu"
    if isinstance(value, str):
        value = torch.device(value)
    return value


__DEFAULT_GLOBALS = {
    "cache": {
        "values": {
            "user": None,
            "env": "AAC_METRICS_CACHE_PATH",
            "package": osp.join("~", ".cache"),
        },
        "process": __process_path,
    },
    "device": {
        "values": {
            "env": "AAC_METRICS_DEVICE",
            "package": "auto",
        },
        "process": __process_device,
    },
    "java": {
        "values": {
            "user": None,
            "env": "AAC_METRICS_JAVA_PATH",
            "package": "java",
        },
        "process": __process_path,
    },
    "tmp": {
        "values": {
            "user": None,
            "env": "AAC_METRICS_TMP_PATH",
            "package": tempfile.gettempdir(),
        },
        "process": __process_path,
    },
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Optional, Union

import torch

from sentence_transformers import SentenceTransformer
from torch import Tensor
from transformers.models.auto.tokenization_auto import AutoTokenizer

from aac_metrics.functional.fer import (
    fer,
    _load_echecker_and_tokenizer,
    BERTFlatClassifier,
)
from aac_metrics.functional.sbert_sim import sbert_sim, _load_sbert
from aac_metrics.utils.checks import check_metric_inputs


pylog = logging.getLogger(__name__)


def fense(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    # SBERT args
    sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
    # FluencyError args
    echecker: Union[str, BERTFlatClassifier] = "echecker_clotho_audiocaps_base",
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    error_threshold: float = 0.9,
    device: Union[str, torch.device, None] = "auto",
    batch_size: int = 32,
    reset_state: bool = True,
    return_probs: bool = False,
    # Other args
    penalty: float = 0.9,
    verbose: int = 0,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Tensor]]]:
    """Fluency ENhanced Sentence-bert Evaluation (FENSE)

    - Paper: https://arxiv.org/abs/2110.04684
    - Original implementation: https://github.com/blmoistawinde/fense

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param sbert_model: The sentence BERT model used to extract sentence embeddings for cosine-similarity. defaults to "paraphrase-TinyBERT-L6-v2".
    :param echecker: The echecker model used to detect fluency errors.
        Can be "echecker_clotho_audiocaps_base", "echecker_clotho_audiocaps_tiny", "none" or None.
        defaults to "echecker_clotho_audiocaps_base".
    :param echecker_tokenizer: The tokenizer of the echecker model.
        If None and echecker is not None, this value will be inferred with `echecker.model_type`.
        defaults to None.
    :param error_threshold: The threshold used to detect fluency errors for echecker model. defaults to 0.9.
    :param penalty: The penalty coefficient applied. Higher value means to lower the cos-sim scores when an error is detected. defaults to 0.9.
    :param device: The PyTorch device used to run FENSE models. If "auto", it will use cuda if available. defaults to "auto".
    :param batch_size: The batch size of the sBERT and echecker models. defaults to 32.
    :param reset_state: If True, reset the state of the PyTorch global generator after the initialization of the pre-trained models. defaults to True.
    :param return_probs: If True, return each individual error probability given by the fluency detector model. defaults to False.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    check_metric_inputs(candidates, mult_references)

    # Init models
    sbert_model, echecker, echecker_tokenizer = _load_models_and_tokenizer(
        sbert_model=sbert_model,
        echecker=echecker,
        echecker_tokenizer=echecker_tokenizer,
        device=device,
        reset_state=reset_state,
        verbose=verbose,
    )
    sbert_sim_outs: tuple[dict[str, Tensor], dict[str, Tensor]] = sbert_sim(  # type: ignore
        candidates=candidates,
        mult_references=mult_references,
        return_all_scores=True,
        sbert_model=sbert_model,
        device=device,
        batch_size=batch_size,
        reset_state=reset_state,
        verbose=verbose,
    )
    fer_outs: tuple[dict[str, Tensor], dict[str, Tensor]] = fer(  # type: ignore
        candidates=candidates,
        return_all_scores=True,
        echecker=echecker,
        echecker_tokenizer=echecker_tokenizer,
        error_threshold=error_threshold,
        device=device,
        batch_size=batch_size,
        reset_state=reset_state,
        return_probs=return_probs,
        verbose=verbose,
    )
    fense_outs = _fense_from_outputs(sbert_sim_outs, fer_outs, penalty)

    if return_all_scores:
        return fense_outs
    else:
        return fense_outs[0]["fense"]


def _fense_from_outputs(
    sbert_sim_outs: tuple[dict[str, Tensor], dict[str, Tensor]],
    fer_outs: tuple[dict[str, Tensor], dict[str, Tensor]],
    penalty: float = 0.9,
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Combines SBERT and FER outputs.

    Based on https://github.com/blmoistawinde/fense/blob/main/fense/evaluator.py#L121
    """
    sbert_sim_outs_corpus, sbert_sim_outs_sents = sbert_sim_outs
    fer_outs_corpus, fer_outs_sents = fer_outs

    sbert_sims_scores = sbert_sim_outs_sents["sbert_sim"]
    fer_scores = fer_outs_sents["fer"]
    fense_scores = sbert_sims_scores * (1.0 - penalty * fer_scores)
    fense_score = torch.as_tensor(
        fense_scores.cpu()
        .numpy()
        .mean(),  # note: use numpy mean to keep the same values than the original fense
        device=fense_scores.device,
    )

    fense_outs_corpus = sbert_sim_outs_corpus | fer_outs_corpus | {"fense": fense_score}
    fense_outs_sents = sbert_sim_outs_sents | fer_outs_sents | {"fense": fense_scores}
    fense_outs = fense_outs_corpus, fense_outs_sents

    return fense_outs


def _load_models_and_tokenizer(
    sbert_model: Union[str, SentenceTransformer] = "paraphrase-TinyBERT-L6-v2",
    echecker: Union[str, BERTFlatClassifier] = "echecker_clotho_audiocaps_base",
    echecker_tokenizer: Optional[AutoTokenizer] = None,
    device: Union[str, torch.device, None] = "auto",
    reset_state: bool = True,
    verbose: int = 0,
) -> tuple[SentenceTransformer, BERTFlatClassifier, AutoTokenizer]:
    sbert_model = _load_sbert(
        sbert_model=sbert_model, device=device, reset_state=reset_state
    )
    echecker, echecker_tokenizer = _load_echecker_and_tokenizer(
        echecker=echecker,
        echecker_tokenizer=echecker_tokenizer,
        device=device,
        reset_state=reset_state,
        verbose=verbose,
    )
    return sbert_model, echecker, echecker_tokenizer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.bleu import (
    BLEU_OPTIONS,
    _bleu_compute,
    _bleu_update,
)


class BLEU(AACMetric[Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]]):
    """BiLingual Evaluation Understudy metric class.

    - Paper: https://www.aclweb.org/anthology/P02-1040.pdf

    For more information, see :func:`~aac_metrics.functional.bleu.bleu`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        n: int = 4,
        option: str = "closest",
        verbose: int = 0,
        tokenizer: Callable[[str], list[str]] = str.split,
    ) -> None:
        if option not in BLEU_OPTIONS:
            raise ValueError(
                f"Invalid option {option=}. (expected one of {BLEU_OPTIONS})"
            )

        super().__init__()
        self._return_all_scores = return_all_scores
        self._n = n
        self._option = option
        self._verbose = verbose
        self._tokenizer = tokenizer

        self._cooked_cands = []
        self._cooked_mrefs = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return _bleu_compute(
            cooked_cands=self._cooked_cands,
            cooked_mrefs=self._cooked_mrefs,
            return_all_scores=self._return_all_scores,
            n=self._n,
            option=self._option,
            verbose=self._verbose,
            return_1_to_n=False,
        )

    def extra_repr(self) -> str:
        hparams = {"n": self._n}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return (f"bleu_{self._n}",)

    def reset(self) -> None:
        self._cooked_cands = []
        self._cooked_mrefs = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._cooked_cands, self._cooked_mrefs = _bleu_update(
            candidates,
            mult_references,
            self._n,
            self._tokenizer,
            self._cooked_cands,
            self._cooked_mrefs,
        )


class BLEU1(BLEU):
    def __init__(
        self,
        return_all_scores: bool = True,
        option: str = "closest",
        verbose: int = 0,
        tokenizer: Callable[[str], list[str]] = str.split,
    ) -> None:
        super().__init__(
            return_all_scores=return_all_scores,
            n=1,
            option=option,
            verbose=verbose,
            tokenizer=tokenizer,
        )


class BLEU2(BLEU):
    def __init__(
        self,
        return_all_scores: bool = True,
        option: str = "closest",
        verbose: int = 0,
        tokenizer: Callable[[str], list[str]] = str.split,
    ) -> None:
        super().__init__(
            return_all_scores=return_all_scores,
            n=2,
            option=option,
            verbose=verbose,
            tokenizer=tokenizer,
        )


class BLEU3(BLEU):
    def __init__(
        self,
        return_all_scores: bool = True,
        option: str = "closest",
        verbose: int = 0,
        tokenizer: Callable[[str], list[str]] = str.split,
    ) -> None:
        super().__init__(
            return_all_scores=return_all_scores,
            n=3,
            option=option,
            verbose=verbose,
            tokenizer=tokenizer,
        )


class BLEU4(BLEU):
    def __init__(
        self,
        return_all_scores: bool = True,
        option: str = "closest",
        verbose: int = 0,
        tokenizer: Callable[[str], list[str]] = str.split,
    ) -> None:
        super().__init__(
            return_all_scores=return_all_scores,
            n=4,
            option=option,
            verbose=verbose,
            tokenizer=tokenizer,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib
import os.path as osp
import platform
import shutil
import subprocess
import sys
import unittest

from pathlib import Path
from typing import Callable, Dict, List, Tuple
from unittest import TestCase

from torch import Tensor

from aac_metrics.functional.evaluate import evaluate
from aac_metrics.eval import load_csv_file
from aac_metrics.utils.globals import (
    get_default_tmp_path,
)
from aac_metrics.download import _download_spice


class TestCompareCaptionEvaluationTools(TestCase):
    # Note: "cet" is here an acronym for "caption evaluation tools"

    # Set Up methods
    @classmethod
    def setUpClass(cls) -> None:
        cls.evaluate_metrics_from_lists = cls._import_cet_eval_func()

    @classmethod
    def _import_cet_eval_func(
        cls,
    ) -> Callable[
        [List[str], List[List[str]]],
        Tuple[Dict[str, float], Dict[int, Dict[str, float]]],
    ]:
        cet_path = osp.join(osp.dirname(__file__), "caption-evaluation-tools")
        on_windows = platform.system() == "Windows"

        cet_cache_path = Path(
            cet_path,
            "coco_caption",
            "pycocoevalcap",
        )
        stanford_fpath = cet_cache_path.joinpath(
            "spice",
            "lib",
            "stanford-corenlp-3.6.0.jar",
        )
        if not osp.isfile(stanford_fpath):
            if not on_windows:
                # Use CET installation
                command = ["bash", "get_stanford_models.sh"]
                subprocess.check_call(
                    command,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    cwd=osp.join(cet_path, "coco_caption"),
                    shell=on_windows,
                )
            else:
                # Use aac-metrics SPICE installation, but it requires to move some files after
                _download_spice(str(cet_cache_path), clean_archives=True, verbose=2)
                shutil.copytree(
                    cet_cache_path.joinpath("aac-metrics", "spice"),
                    cet_cache_path.joinpath("spice"),
                    dirs_exist_ok=True,
                )
                shutil.rmtree(cet_cache_path.joinpath("aac-metrics"))

        # Append cet_path to allow imports of "caption" in eval_metrics.py.
        sys.path.append(cet_path)
        # Override cache and tmp dir to avoid outputs in source code.
        spice_module = importlib.import_module("coco_caption.pycocoevalcap.spice.spice")
        spice_module.CACHE_DIR = get_default_tmp_path()  # type: ignore
        spice_module.TEMP_DIR = get_default_tmp_path()  # type: ignore
        eval_metrics_module = importlib.import_module("eval_metrics")
        evaluate_metrics_from_lists = eval_metrics_module.evaluate_metrics_from_lists
        return evaluate_metrics_from_lists

    # Tests methods
    def test_example_0(self) -> None:
        cands, mrefs = self._get_example_0()
        self._test_with_example(cands, mrefs)

    def test_example_1(self) -> None:
        fpath = Path(__file__).parent.parent.joinpath("data", "example_1.csv")
        cands, mrefs = load_csv_file(fpath)
        self._test_with_example(cands, mrefs)

    def test_example_2(self) -> None:
        fpath = Path(__file__).parent.parent.joinpath("data", "example_2.csv")
        cands, mrefs = load_csv_file(fpath)
        self._test_with_example(cands, mrefs)

    # Others methods
    def _get_example_0(self) -> tuple[list[str], list[list[str]]]:
        cands = [
            "a man is speaking",
            "birds chirping",
            "rain is falling in the background",
        ]
        mrefs = [
            [
                "man speaks",
                "man is speaking",
                "a man speaks",
                "man talks",
                "someone is talking",
            ],
            ["a bird is chirping"] * 5,
            ["heavy rain noise"] * 5,
        ]
        return cands, mrefs

    def _test_with_example(self, cands: list[str], mrefs: list[list[str]]) -> None:
        if platform.system() == "Windows":
            return None

        corpus_scores, _ = evaluate(cands, mrefs, metrics="dcase2020", preprocess=True)

        self.assertIsInstance(corpus_scores, dict)

        for name, score in corpus_scores.items():
            self.assertIsInstance(score, Tensor, f"Invalid score type for {name=}")
            self.assertEqual(score.ndim, 0, f"Invalid score ndim for {name=}")

        cet_outs = self.__class__.evaluate_metrics_from_lists(cands, mrefs)
        cet_global_scores, _cet_sents_scores = cet_outs

        cet_global_scores = {k.lower(): v for k, v in cet_global_scores.items()}
        cet_global_scores = {
            (k if k != "cider" else "cider_d"): v for k, v in cet_global_scores.items()
        }

        self.assertIsInstance(cet_global_scores, dict)
        self.assertListEqual(list(corpus_scores.keys()), list(cet_global_scores.keys()))

        for metric_name, v1 in corpus_scores.items():
            v1 = v1.item()
            v2 = cet_global_scores[metric_name]
            self.assertEqual(v1, v2, f"{metric_name=}")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .bert_score_mrefs import bert_score_mrefs
from .bleu import bleu, bleu_1, bleu_2, bleu_3, bleu_4
from .cider_d import cider_d
from .evaluate import dcase2023_evaluate, evaluate
from .fense import fense
from .fer import fer
from .meteor import meteor
from .rouge_l import rouge_l
from .sbert_sim import sbert_sim
from .spice import spice
from .spider import spider
from .spider_fl import spider_fl
from .spider_max import spider_max
from .vocab import vocab


__all__ = [
    "bert_score_mrefs",
    "bleu",
    "bleu_1",
    "bleu_2",
    "bleu_3",
    "bleu_4",
    "cider_d",
    "dcase2023_evaluate",
    "evaluate",
    "fense",
    "fer",
    "meteor",
    "rouge_l",
    "sbert_sim",
    "spice",
    "spider",
    "spider_fl",
    "spider_max",
    "vocab",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import unittest

from unittest import TestCase

from aac_metrics.functional.evaluate import evaluate


class TestAll(TestCase):
    def test_example_1(self) -> None:
        cands: list[str] = ["a man is speaking", "rain falls"]
        mrefs: list[list[str]] = [
            [
                "a man speaks.",
                "someone speaks.",
                "a man is speaking while a bird is chirping in the background",
            ],
            ["rain is falling hard on a surface"],
        ]

        _ = evaluate(cands, mrefs, metrics="all")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import unittest

from unittest import TestCase

from aac_metrics.utils.checks import (
    is_mono_sents,
    is_mult_sents,
    _check_java_version,
    MIN_JAVA_MAJOR_VERSION,
    MAX_JAVA_MAJOR_VERSION,
)
from aac_metrics.utils.collections import flat_list, unflat_list


class TestUtils(TestCase):
    # Tests methods
    def test_misc_functions_1(self) -> None:
        lst = [
            list(map(str, range(random.randint(0, 100))))
            for _ in range(random.randint(0, 10))
        ]
        for sublst in lst:
            random.shuffle(sublst)
        random.shuffle(lst)

        self.assertTrue(is_mult_sents(lst))

        flatten, sizes = flat_list(lst)

        self.assertTrue(is_mono_sents(flatten))
        self.assertEqual(len(lst), len(sizes))
        self.assertEqual(len(flatten), sum(sizes))

        unflat = unflat_list(flatten, sizes)

        self.assertTrue(is_mult_sents(unflat))
        self.assertEqual(len(lst), len(unflat))
        self.assertListEqual(lst, unflat)

    def test_check_java_versions(self) -> None:
        test_set = [
            ("1.0.0", False),
            ("1.7.0", False),
            ("1.8.0", True),
            ("1.9.0", False),
            ("1.10.0", False),
            ("9.0.0", True),
            ("10.0.0", True),
            ("11.0.0", True),
            ("12.0.0", True),
            ("13.0.0", True),
            ("14.0.0", False),
            ("17.0.0", False),
            ("20.0.0", False),
        ]
        for version, expected in test_set:
            output = _check_java_version(
                version, MIN_JAVA_MAJOR_VERSION, MAX_JAVA_MAJOR_VERSION
            )
            self.assertEqual(output, expected, f"{version=}")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .bert_score_mrefs import BERTScoreMRefs
from .bleu import BLEU, BLEU1, BLEU2, BLEU3, BLEU4
from .cider_d import CIDErD
from .evaluate import DCASE2023Evaluate, Evaluate
from .fense import FENSE
from .fer import FER
from .meteor import METEOR
from .rouge_l import ROUGEL
from .sbert_sim import SBERTSim
from .spice import SPICE
from .spider import SPIDEr
from .spider_fl import SPIDErFL
from .spider_max import SPIDErMax
from .vocab import Vocab


__all__ = [
    "BERTScoreMRefs",
    "BLEU",
    "BLEU1",
    "BLEU2",
    "BLEU3",
    "BLEU4",
    "CIDErD",
    "DCASE2023Evaluate",
    "Evaluate",
    "FENSE",
    "FER",
    "METEOR",
    "ROUGEL",
    "SBERTSim",
    "SPICE",
    "SPIDEr",
    "SPIDErFL",
    "SPIDErMax",
    "Vocab",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os.path as osp
import platform
import subprocess

from pathlib import Path
from subprocess import Popen
from typing import Iterable, Optional, Union

import torch

from torch import Tensor

from aac_metrics.utils.checks import check_java_path, check_metric_inputs
from aac_metrics.utils.globals import _get_cache_path, _get_java_path


pylog = logging.getLogger(__name__)


DNAME_METEOR_CACHE = osp.join("aac-metrics", "meteor")
FNAME_METEOR_JAR = osp.join(DNAME_METEOR_CACHE, "meteor-1.5.jar")
SUPPORTED_LANGUAGES = ("en", "cz", "de", "es", "fr")


def meteor(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    java_max_memory: str = "2G",
    language: str = "en",
    use_shell: Optional[bool] = None,
    params: Optional[Iterable[float]] = None,
    weights: Optional[Iterable[float]] = None,
    verbose: int = 0,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """Metric for Evaluation of Translation with Explicit ORdering function.

    - Paper: https://dl.acm.org/doi/pdf/10.5555/1626355.1626389
    - Documentation: https://www.cs.cmu.edu/~alavie/METEOR/README.html
    - Original implementation: https://github.com/tylin/coco-caption

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param java_max_memory: The maximal java memory used. defaults to "2G".
    :param language: The language used for stem, synonym and paraphrase matching.
        Can be one of ("en", "cz", "de", "es", "fr").
        defaults to "en".
    :param use_shell: Optional argument to force use os-specific shell for the java subprogram.
        If None, it will use shell only on Windows OS.
        defaults to None.
    :param params: List of 4 parameters (alpha, beta gamma delta) used in METEOR metric.
        If None, it will use the default of the java program, which is (0.85, 0.2, 0.6, 0.75).
        defaults to None.
    :param weights: List of 4 parameters (w1, w2, w3, w4) used in METEOR metric.
        If None, it will use the default of the java program, which is (1.0 1.0 0.6 0.8).
        defaults to None.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    check_metric_inputs(candidates, mult_references)

    cache_path = _get_cache_path(cache_path)
    java_path = _get_java_path(java_path)

    meteor_jar_fpath = osp.join(cache_path, FNAME_METEOR_JAR)

    if use_shell is None:
        use_shell = platform.system() == "Windows"

    if __debug__:
        if not osp.isfile(meteor_jar_fpath):
            raise FileNotFoundError(
                f"Cannot find JAR file '{meteor_jar_fpath}' for METEOR metric. Maybe run 'aac-metrics-download' or specify another 'cache_path' directory."
            )
        if not check_java_path(java_path):
            raise RuntimeError(
                f"Invalid Java executable to compute METEOR score. ({java_path})"
            )

    if language not in SUPPORTED_LANGUAGES:
        raise ValueError(
            f"Invalid argument {language=}. (expected one of {SUPPORTED_LANGUAGES})"
        )

    # Note: override localization to avoid errors due to double conversion (https://github.com/Labbeti/aac-metrics/issues/9)
    meteor_cmd = [
        java_path,
        "-Duser.country=US",
        "-Duser.language=en",
        "-jar",
        f"-Xmx{java_max_memory}",
        meteor_jar_fpath,
        "-",
        "-",
        "-stdio",
        "-l",
        language,
        "-norm",
    ]

    if params is not None:
        params = list(params)
        if len(params) != 4:
            raise ValueError(
                f"Invalid argument {params=}. (expected 4 params but found {len(params)})"
            )
        params_arg = " ".join(map(str, params))
        meteor_cmd += ["-p", f"{params_arg}"]

    if weights is not None:
        weights = list(weights)
        if len(weights) != 4:
            raise ValueError(
                f"Invalid argument {weights=}. (expected 4 params but found {len(weights)})"
            )
        weights_arg = " ".join(map(str, weights))
        meteor_cmd += ["-w", f"{weights_arg}"]

    if verbose >= 2:
        pylog.debug(
            f"Run METEOR java code with: {' '.join(meteor_cmd)} and {use_shell=}"
        )

    meteor_process = Popen(
        meteor_cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=use_shell,
    )

    n_candidates = len(candidates)
    encoded_cands_and_mrefs = [
        _encode_cand_and_refs(cand, refs)
        for cand, refs in zip(candidates, mult_references)
    ]
    del candidates, mult_references

    # Encode candidates and references
    eval_line = "EVAL"
    for encoded in encoded_cands_and_mrefs:
        assert meteor_process.stdin is not None, "INTERNAL METEOR process error"
        meteor_process.stdin.write(encoded)
        meteor_process.stdin.flush()
        assert meteor_process.stdout is not None, "INTERNAL METEOR process error"
        stat = meteor_process.stdout.readline().decode().strip()
        eval_line += " ||| {}".format(stat)

    # Eval encoded candidates and references
    assert meteor_process.stdin is not None, "INTERNAL METEOR process error"
    if verbose >= 3:
        pylog.debug(f"Write line {eval_line=}.")

    process_inputs = "{}\n".format(eval_line).encode()
    meteor_process.stdin.write(process_inputs)
    meteor_process.stdin.flush()

    # Read scores
    assert meteor_process.stdout is not None, "INTERNAL METEOR process error"
    meteor_scores = []
    for i in range(n_candidates):
        process_out_i = meteor_process.stdout.readline().strip()
        try:
            meteor_scores_i = float(process_out_i)
        except ValueError as err:
            pylog.error(
                f"Invalid METEOR stdout. (cannot convert sentence score to float {process_out_i=} with {i=})"
            )
            raise err
        meteor_scores.append(meteor_scores_i)

    process_out = meteor_process.stdout.readline().strip()
    try:
        meteor_score = float(process_out)
    except ValueError as err:
        pylog.error(
            f"Invalid METEOR stdout. (cannot convert global score to float {process_out=})"
        )
        raise err

    meteor_process.stdin.close()
    meteor_process.kill()
    meteor_process.wait()

    dtype = torch.float64
    meteor_score = torch.as_tensor(meteor_score, dtype=dtype)
    meteor_scores = torch.as_tensor(meteor_scores, dtype=dtype)

    if return_all_scores:
        meteor_outs_corpus = {
            "meteor": meteor_score,
        }
        meteor_outs_sents = {
            "meteor": meteor_scores,
        }
        meteor_outs = meteor_outs_corpus, meteor_outs_sents
        return meteor_outs
    else:
        return meteor_score


def _encode_cand_and_refs(candidate: str, references: list[str]) -> bytes:
    # SCORE ||| reference 1 words ||| ... ||| reference N words ||| candidate words
    candidate = candidate.replace("|||", "").replace("  ", " ")
    score_line = " ||| ".join(("SCORE", " ||| ".join(references), candidate))
    encoded = "{}\n".format(score_line).encode()
    return encoded



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Optional, Union

import torch

from torch import nn, Tensor
from torchmetrics.functional.text.bert import bert_score, _DEFAULT_MODEL
from transformers.models.auto.modeling_auto import AutoModel
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers import logging as tfmers_logging

from aac_metrics.utils.checks import check_metric_inputs
from aac_metrics.utils.collections import flat_list, unflat_list, duplicate_list
from aac_metrics.utils.globals import _get_device


def bert_score_mrefs(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    model: Union[str, nn.Module] = _DEFAULT_MODEL,
    tokenizer: Optional[Callable] = None,
    device: Union[str, torch.device, None] = "auto",
    batch_size: int = 32,
    num_threads: int = 0,
    max_length: int = 64,
    reset_state: bool = True,
    idf: bool = False,
    reduction: str = "max",
    filter_nan: bool = True,
    verbose: int = 0,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """BERTScore metric which supports multiple references.

    The implementation is based on the bert_score implementation of torchmetrics.

    - Paper: https://arxiv.org/pdf/1904.09675.pdf

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param model: The model name or the instantiated model to use to compute token embeddings.
        defaults to "roberta-large".
    :param tokenizer: The fast tokenizer used to split sentences into words.
        If None, use the tokenizer corresponding to the model argument.
        defaults to None.
    :param device: The PyTorch device used to run the BERT model. defaults to "auto".
    :param batch_size: The batch size used in the model forward.
    :param num_threads: A number of threads to use for a dataloader. defaults to 0.
    :param max_length: Max length when encoding sentences to tensor ids. defaults to 64.
    :param idf: Whether or not using Inverse document frequency to ponderate the BERTScores. defaults to False.
    :param reduction: The reduction function to apply between multiple references for each audio. defaults to "mean".
    :param filter_nan: If True, replace NaN scores by 0.0. defaults to True.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    check_metric_inputs(candidates, mult_references)

    if isinstance(model, str):
        if tokenizer is not None:
            raise ValueError(
                f"Invalid argument combinaison {model=} with {tokenizer=}."
            )

        model, tokenizer = _load_model_and_tokenizer(
            model=model,
            tokenizer=tokenizer,
            device=device,
            reset_state=reset_state,
            verbose=verbose,
        )

    elif isinstance(model, nn.Module):
        if tokenizer is None:
            raise ValueError(
                f"Invalid argument combinaison {model=} with {tokenizer=}."
            )

    else:
        raise ValueError(
            f"Invalid argument type {type(model)=}. (expected str or nn.Module)"
        )

    device = _get_device(device)
    flat_mrefs, sizes = flat_list(mult_references)
    duplicated_cands = duplicate_list(candidates, sizes)
    assert len(duplicated_cands) == len(flat_mrefs)

    tfmers_verbosity = tfmers_logging.get_verbosity()
    if verbose <= 1:
        tfmers_logging.set_verbosity_error()

    sents_scores = bert_score(
        preds=duplicated_cands,
        target=flat_mrefs,
        model_name_or_path=None,
        model=model,  # type: ignore
        user_tokenizer=tokenizer,
        device=device,
        batch_size=batch_size,
        num_threads=num_threads,
        verbose=verbose >= 3,
        max_length=max_length,
        idf=idf,
    )
    if verbose <= 1:
        # Restore previous verbosity level
        tfmers_logging.set_verbosity(tfmers_verbosity)

    # note: torchmetrics returns a float if input contains 1 cand and 1 ref, even in list
    if len(duplicated_cands) == 1 and all(
        isinstance(v, float) for v in sents_scores.values()
    ):
        sents_scores = {k: [v] for k, v in sents_scores.items()}

    # sents_scores keys: "precision", "recall", "f1"
    sents_scores = {k: unflat_list(v, sizes) for k, v in sents_scores.items()}  # type: ignore

    if not return_all_scores:
        sents_scores = {"f1": sents_scores["f1"]}

    dtype = torch.float32

    if reduction == "mean":
        reduction_fn = torch.mean
    elif reduction == "max":
        reduction_fn = _max_reduce
    elif reduction == "min":
        reduction_fn = _min_reduce
    else:
        REDUCTIONS = ("mean", "max", "min")
        raise ValueError(
            f"Invalid argument {reduction=}. (expected one of {REDUCTIONS})"
        )

    if len(sizes) > 0 and all(size == sizes[0] for size in sizes):
        sents_scores = {
            k: reduction_fn(torch.as_tensor(v, dtype=dtype), dim=1)
            for k, v in sents_scores.items()
        }
    else:
        sents_scores = {
            k: torch.stack([reduction_fn(torch.as_tensor(vi, dtype=dtype)) for vi in v])
            for k, v in sents_scores.items()
        }

    sents_scores = {f"bert_score.{k}": v for k, v in sents_scores.items()}

    if filter_nan:
        # avoid NaN that can occur in some cases
        sents_scores = {
            k: v.masked_fill(v.isnan(), 0.0) for k, v in sents_scores.items()
        }

    corpus_scores = {k: v.mean() for k, v in sents_scores.items()}

    if return_all_scores:
        return corpus_scores, sents_scores
    else:
        return corpus_scores["bert_score.f1"]


def _load_model_and_tokenizer(
    model: Union[str, nn.Module],
    tokenizer: Optional[Callable],
    device: Union[str, torch.device, None],
    reset_state: bool,
    verbose: int,
) -> tuple[nn.Module, Optional[Callable]]:
    state = torch.random.get_rng_state()

    device = _get_device(device)
    if isinstance(model, str):
        tfmers_verbosity = tfmers_logging.get_verbosity()
        if verbose <= 1:
            tfmers_logging.set_verbosity_error()

        # WARNING: tokenizer must be initialized BEFORE model to avoid connection errors
        tokenizer = AutoTokenizer.from_pretrained(model)
        model = AutoModel.from_pretrained(model)  # type: ignore

        if verbose <= 1:
            # Restore previous verbosity level
            tfmers_logging.set_verbosity(tfmers_verbosity)

    model.eval()  # type: ignore
    model.to(device=device)  # type: ignore

    if reset_state:
        torch.random.set_rng_state(state)

    return model, tokenizer  # type: ignore


def _max_reduce(x: Tensor, dim: Optional[int] = None) -> Tensor:
    if dim is None:
        return x.max()
    else:
        return x.max(dim=dim).values


def _min_reduce(x: Tensor, dim: Optional[int] = None) -> Tensor:
    if dim is None:
        return x.min()
    else:
        return x.min(dim=dim).values



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Iterable, Optional, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.meteor import meteor


class METEOR(AACMetric[Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]]):
    """Metric for Evaluation of Translation with Explicit ORdering metric class.

    - Paper: https://dl.acm.org/doi/pdf/10.5555/1626355.1626389
    - Documentation: https://www.cs.cmu.edu/~alavie/METEOR/README.html

    For more information, see :func:`~aac_metrics.functional.meteor.meteor`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        cache_path: Union[str, Path, None] = None,
        java_path: Union[str, Path, None] = None,
        java_max_memory: str = "2G",
        language: str = "en",
        use_shell: Optional[bool] = None,
        params: Optional[Iterable[float]] = None,
        weights: Optional[Iterable[float]] = None,
        verbose: int = 0,
    ) -> None:
        super().__init__()
        self._return_all_scores = return_all_scores
        self._cache_path = cache_path
        self._java_path = java_path
        self._java_max_memory = java_max_memory
        self._language = language
        self._use_shell = use_shell
        self._params = params
        self._weights = weights
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return meteor(
            candidates=self._candidates,
            mult_references=self._mult_references,
            return_all_scores=self._return_all_scores,
            cache_path=self._cache_path,
            java_path=self._java_path,
            java_max_memory=self._java_max_memory,
            language=self._language,
            use_shell=self._use_shell,
            params=self._params,
            weights=self._weights,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        hparams = {"java_max_memory": self._java_max_memory, "language": self._language}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return ("meteor",)

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._candidates += candidates
        self._mult_references += mult_references



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import math

from typing import Callable, Union

import torch

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.vocab import vocab


pylog = logging.getLogger(__name__)


class Vocab(AACMetric[Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]]):
    """VocabStats class.

    For more information, see :func:`~aac_metrics.functional.vocab.vocab`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = math.inf

    def __init__(
        self,
        return_all_scores: bool = True,
        seed: Union[None, int, torch.Generator] = 1234,
        tokenizer: Callable[[str], list[str]] = str.split,
        dtype: torch.dtype = torch.float64,
        pop_strategy: str = "max",
        verbose: int = 0,
    ) -> None:
        super().__init__()
        self._return_all_scores = return_all_scores
        self._seed = seed
        self._tokenizer = tokenizer
        self._dtype = dtype
        self._pop_strategy = pop_strategy
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return vocab(
            candidates=self._candidates,
            mult_references=self._mult_references,
            return_all_scores=self._return_all_scores,
            seed=self._seed,
            tokenizer=self._tokenizer,
            dtype=self._dtype,
            pop_strategy=self._pop_strategy,
            verbose=self._verbose,
        )

    def get_output_names(self) -> tuple[str, ...]:
        return (
            "vocab",
            "vocab.mrefs_full",
            "vocab.ratio_full",
            "vocab.mrefs_avg",
            "vocab.mrefs_std",
            "vocab.ratio_avg",
        )

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: Union[list[list[str]], None] = None,
    ) -> None:
        self._candidates += candidates

        if mult_references is not None:
            if self._mult_references is None:
                self._mult_references = []
            else:
                self._mult_references += mult_references
        else:
            self._mult_references = None

        if self._mult_references is not None and len(self._candidates) != len(
            self._mult_references
        ):
            raise ValueError(
                f"Invalid number of sentences for {self.__class__.__name__}. (found {len(candidates)} candidates and {len(self._mult_references)} references)"
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import sys

from pathlib import Path
from typing import Dict

import torch
import yaml

import aac_metrics

from aac_metrics.utils.checks import _get_java_version
from aac_metrics.utils.globals import (
    get_default_cache_path,
    get_default_java_path,
    get_default_tmp_path,
)


def get_package_repository_path() -> str:
    """Return the absolute path where the source code of this package is installed."""
    return str(Path(__file__).parent.parent.parent)


def get_java_version() -> str:
    try:
        java_version = _get_java_version(get_default_java_path())
        return java_version
    except ValueError:
        return "UNKNOWN"


def get_install_info() -> Dict[str, str]:
    """Return a dictionary containing the version python, the os name, the architecture name and the versions of the following packages: aac_datasets, torch, torchaudio."""

    return {
        "aac_metrics": aac_metrics.__version__,
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "os": platform.system(),
        "architecture": platform.architecture()[0],
        "torch": str(torch.__version__),
        "package_path": get_package_repository_path(),
        "cache_path": get_default_cache_path(),
        "java_path": get_default_java_path(),
        "tmp_path": get_default_tmp_path(),
        "java_version": get_java_version(),
    }


def print_install_info() -> None:
    """Show main packages versions."""
    install_info = get_install_info()
    print(yaml.dump(install_info, sort_keys=False))


if __name__ == "__main__":
    print_install_info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Callable, Union

import torch

from torch import Tensor

from aac_metrics.utils.checks import check_metric_inputs, is_mono_sents


pylog = logging.getLogger(__name__)


def vocab(
    candidates: list[str],
    mult_references: Union[list[list[str]], None],
    return_all_scores: bool = True,
    seed: Union[None, int, torch.Generator] = 1234,
    tokenizer: Callable[[str], list[str]] = str.split,
    dtype: torch.dtype = torch.float64,
    pop_strategy: str = "max",
    verbose: int = 0,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """Compute vocabulary statistics.

    Returns the candidate corpus vocabulary length, the references vocabulary length, the average vocabulary length for single references, and the vocabulary ratios between candidates and references.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target. Can also be None.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param seed: Random seed used to compute average vocabulary length for multiple references. defaults to 1234.
    :param tokenizer: The function used to split a sentence into tokens. defaults to str.split.
    :param dtype: Torch floating point dtype for numerical precision. defaults to torch.float64.
    :param pop_strategy: Strategy to compute average reference vocab. defaults to "max".
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    if mult_references is not None:
        check_metric_inputs(candidates, mult_references)
    elif not is_mono_sents(candidates):
        error_msg = f"Invalid candidates type. (expected list[str], found {candidates.__class__.__name__})"
        raise ValueError(error_msg)

    tok_cands = list(map(tokenizer, candidates))
    del candidates

    vocab_cands_len = _corpus_vocab(tok_cands, dtype)
    if not return_all_scores:
        return vocab_cands_len

    sents_scores = {}
    corpus_scores = {
        "vocab.cands": vocab_cands_len,
    }

    if mult_references is not None:
        if len(mult_references) <= 0:
            raise ValueError(
                f"Invalid number of references. (found {len(mult_references)} references)"
            )
        tok_mrefs = [list(map(tokenizer, refs)) for refs in mult_references]
        del mult_references

        vocab_mrefs_len_full = _corpus_vocab(
            [ref for refs in tok_mrefs for ref in refs], dtype
        )
        vocab_ratio_len_full = vocab_cands_len / vocab_mrefs_len_full

        if isinstance(seed, int):
            generator = torch.Generator().manual_seed(seed)
        else:
            generator = seed

        if pop_strategy == "max":
            n_samples = max(len(refs) for refs in tok_mrefs)
        elif pop_strategy == "min":
            n_samples = min(len(refs) for refs in tok_mrefs)
        elif isinstance(pop_strategy, int):
            n_samples = pop_strategy
        else:
            POP_STRATEGIES = ("max", "min")
            raise ValueError(
                f"Invalid argument {pop_strategy=}. (expected one of {POP_STRATEGIES} or an integer value)"
            )

        if verbose >= 2:
            pylog.debug(f"Found {n_samples=} with {pop_strategy=}.")

        vocab_mrefs_lens = torch.empty((n_samples,), dtype=dtype)

        for i in range(n_samples):
            indexes = [
                int(torch.randint(0, len(refs), (), generator=generator).item())
                for refs in tok_mrefs
            ]
            popped_refs = [refs[idx] for idx, refs in zip(indexes, tok_mrefs)]
            vocab_mrefs_len_i = _corpus_vocab(popped_refs, dtype)
            vocab_mrefs_lens[i] = vocab_mrefs_len_i

        vocab_mrefs_avg = vocab_mrefs_lens.mean()
        vocab_len_ratio_avg = vocab_cands_len / vocab_mrefs_avg

        corpus_scores |= {
            "vocab.mrefs_full": vocab_mrefs_len_full,
            "vocab.ratio_full": vocab_ratio_len_full,
            "vocab.mrefs_avg": vocab_mrefs_avg,
            "vocab.ratio_avg": vocab_len_ratio_avg,
        }

    return corpus_scores, sents_scores


def _corpus_vocab(tok_sents: list[list[str]], dtype: torch.dtype) -> Tensor:
    corpus_cands_vocab = set(token for sent in tok_sents for token in sent)
    vocab_len = torch.as_tensor(len(corpus_cands_vocab), dtype=dtype)
    return vocab_len


def _sent_vocab(
    tok_sents: list[list[str]],
    dtype: torch.dtype,
) -> tuple[Tensor, Tensor]:
    sents_cands_vocabs = [set(sent) for sent in tok_sents]
    sent_cands_vocabs_lens = torch.as_tensor(
        list(map(len, sents_cands_vocabs)), dtype=dtype
    )
    sent_cands_vocab_len = sent_cands_vocabs_lens.mean()
    return sent_cands_vocab_len, sent_cands_vocabs_lens



#!/usr/bin/env python
# -*- coding: utf-8 -*-


def _print_usage() -> None:
    print(
        "Command line usage :\n"
        "- Download models and external code               : aac-metrics-download ...\n"
        "- Print scores from candidate and references file : aac-metrics-eval -i [FILEPATH]\n"
        "- Print package version                           : aac-metrics-info\n"
        "- Show this usage page                            : aac-metrics\n"
    )


if __name__ == "__main__":
    _print_usage()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import TypeVar

T = TypeVar("T")


def flat_list(lst: list[list[T]]) -> tuple[list[T], list[int]]:
    """Return a flat version of the input list of sublists with each sublist size."""
    flatten_lst = [element for sublst in lst for element in sublst]
    sizes = [len(sents) for sents in lst]
    return flatten_lst, sizes


def unflat_list(flatten_lst: list[T], sizes: list[int]) -> list[list[T]]:
    """Unflat a list to a list of sublists of given sizes."""
    lst = []
    start = 0
    stop = 0
    for count in sizes:
        stop += count
        lst.append(flatten_lst[start:stop])
        start = stop
    return lst


def duplicate_list(lst: list[T], sizes: list[int]) -> list[T]:
    """Duplicate elements elements of a list with the corresponding sizes.

    Example 1
    ----------
    >>> lst = ["a", "b", "c", "d", "e"]
    >>> sizes = [1, 0, 2, 1, 3]
    >>> duplicate_list(lst, sizes)
    ... ["a", "c", "c", "d", "e", "e", "e"]
    """
    if len(lst) != len(sizes):
        raise ValueError(
            f"Invalid arguments lengths. (found {len(lst)=} != {len(sizes)=})"
        )

    out_size = sum(sizes)
    out: list[T] = [None for _ in range(out_size)]  # type: ignore
    curidx = 0
    for size, elt in zip(sizes, lst):
        out[curidx : curidx + size] = [elt] * size
        curidx += size
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Callable, Union

from torch import Tensor

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.cider_d import (
    _cider_d_compute,
    _cider_d_update,
)


class CIDErD(AACMetric[Union[tuple[dict[str, Tensor], dict[str, Any]], Tensor]]):
    """Consensus-based Image Description Evaluation metric class.

    - Paper: https://arxiv.org/pdf/1411.5726.pdf

    For more information, see :func:`~aac_metrics.functional.cider_d.cider_d`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 10.0

    def __init__(
        self,
        return_all_scores: bool = True,
        n: int = 4,
        sigma: float = 6.0,
        tokenizer: Callable[[str], list[str]] = str.split,
        return_tfidf: bool = False,
        scale: float = 10.0,
    ) -> None:
        super().__init__()
        self._return_all_scores = return_all_scores
        self._n = n
        self._sigma = sigma
        self._tokenizer = tokenizer
        self._return_tfidf = return_tfidf
        self._scale = scale

        self._cooked_cands = []
        self._cooked_mrefs = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return _cider_d_compute(
            cooked_cands=self._cooked_cands,
            cooked_mrefs=self._cooked_mrefs,
            return_all_scores=self._return_all_scores,
            n=self._n,
            sigma=self._sigma,
            return_tfidf=self._return_tfidf,
            scale=self._scale,
        )

    def extra_repr(self) -> str:
        hparams = {"n": self._n, "sigma": self._sigma}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return ("cider_d",)

    def reset(self) -> None:
        self._cooked_cands = []
        self._cooked_mrefs = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._cooked_cands, self._cooked_mrefs = _cider_d_update(
            candidates=candidates,
            mult_references=mult_references,
            n=self._n,
            tokenizer=self._tokenizer,
            prev_cooked_cands=self._cooked_cands,
            prev_cooked_mrefs=self._cooked_mrefs,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from typing import Callable, Union

import numpy as np
import torch

from torch import Tensor

from aac_metrics.utils.checks import check_metric_inputs


pylog = logging.getLogger(__name__)


def rouge_l(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    beta: float = 1.2,
    tokenizer: Callable[[str], list[str]] = str.split,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """Recall-Oriented Understudy for Gisting Evaluation function.

    - Paper: https://aclanthology.org/W04-1013.pdf
    - Original Author: Ramakrishna Vedantam <vrama91@vt.edu>
    - Original implementation: https://github.com/tylin/coco-caption

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param beta: Determines the weight of recall in the combined f-score. defaults to 1.2.
    :param tokenizer: The fast tokenizer used to split sentences into words. defaults to str.split.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    rouge_l_scores = _rouge_l_update(candidates, mult_references, beta, tokenizer, [])
    return _rouge_l_compute(rouge_l_scores, return_all_scores)


def _rouge_l_update(
    candidates: list[str],
    mult_references: list[list[str]],
    beta: float,
    tokenizer: Callable[[str], list[str]],
    prev_rouge_l_scores: list[float],
) -> list[float]:
    check_metric_inputs(candidates, mult_references)

    new_rouge_l_scores = [
        __calc_score(cand, refs, beta, tokenizer)
        for cand, refs in zip(candidates, mult_references)
    ]
    prev_rouge_l_scores += new_rouge_l_scores
    return prev_rouge_l_scores


def _rouge_l_compute(
    rouge_l_scs: list[float],
    return_all_scores: bool,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    # Note: use numpy to compute mean because np.mean and torch.mean can give very small differences
    rouge_l_scores_np = np.array(rouge_l_scs)
    rouge_l_score_np = rouge_l_scores_np.mean()

    rouge_l_score_pt = torch.as_tensor(rouge_l_score_np)
    rouge_l_scores_pt = torch.from_numpy(rouge_l_scores_np)

    if return_all_scores:
        rouge_l_outs_corpus = {
            "rouge_l": rouge_l_score_pt,
        }
        rouge_l_outs_sents = {
            "rouge_l": rouge_l_scores_pt,
        }
        return rouge_l_outs_corpus, rouge_l_outs_sents
    else:
        return rouge_l_score_pt


def __calc_score(
    candidate: str,
    references: list[str],
    beta: float,
    tokenizer: Callable[[str], list[str]] = str.split,
) -> float:
    """Compute ROUGE-L score given one candidate and mult_references for an audio
    :param candidate: list of str : candidate sentence to be evaluated
    :param refs: list of str : Reference sentences for the particular audio to be evaluated
    :returns score: int (ROUGE-L score for the candidate evaluated against mult_references)
    """
    assert len(references) > 0
    prec = []
    rec = []

    # split into tokens
    token_c = tokenizer(candidate)

    # Add Labbeti: returns 0 when candidate is empty.
    if len(token_c) == 0:
        return 0.0

    for reference in references:
        # split into tokens
        token_r = tokenizer(reference)
        # compute the longest common subsequence
        lcs = __my_lcs(token_r, token_c)
        prec.append(lcs / float(len(token_c)))
        rec.append(lcs / float(len(token_r)))

    prec_max = max(prec)
    rec_max = max(rec)

    if prec_max != 0 and rec_max != 0:
        score = ((1 + beta**2) * prec_max * rec_max) / float(
            rec_max + beta**2 * prec_max
        )
    else:
        score = 0.0
    return score


def __my_lcs(string: list[str], sub: list[str]) -> int:
    """
    Calculates longest common subsequence for a pair of tokenized strings
    :param string : list of str : tokens from a string split using whitespace
    :param sub : list of str : shorter string, also split using whitespace
    :returns: length (list of int): length of the longest common subsequence between the two strings
    Note: my_lcs only gives length of the longest common subsequence, not the actual LCS
    """
    if len(string) < len(sub):
        sub, string = string, sub

    lengths = [[0 for i in range(0, len(sub) + 1)] for j in range(0, len(string) + 1)]
    # lengths shape: (len(string)+1, len(sub)+1)

    for j in range(1, len(sub) + 1):
        for i in range(1, len(string) + 1):
            if string[i - 1] == sub[j - 1]:
                lengths[i][j] = lengths[i - 1][j - 1] + 1
            else:
                lengths[i][j] = max(lengths[i - 1][j], lengths[i][j - 1])

    return lengths[len(string)][len(sub)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from unittest import TestCase

import torch

from torch import Tensor

from aac_metrics.classes.spider import SPIDEr
from aac_metrics.classes.spider_max import SPIDErMax


class TestSPIDErMax(TestCase):
    # Tests methods
    def test_sd_vs_sdmax(self) -> None:
        sd = SPIDEr(return_all_scores=False)
        sdmax = SPIDErMax(return_all_scores=False)

        cands, mrefs = self._get_example_0()
        mcands = [[cand] for cand in cands]

        sd_score = sd(cands, mrefs)
        sdmax_score = sdmax(mcands, mrefs)

        assert isinstance(sd_score, Tensor)
        assert isinstance(sdmax_score, Tensor)
        self.assertTrue(
            torch.allclose(sd_score, sdmax_score), f"{sd_score=}, {sdmax_score=}"
        )

    def _get_example_0(self) -> tuple[list[str], list[list[str]]]:
        cands = [
            "a man is speaking",
            "birds chirping",
            "rain is falling in the background",
        ]
        mrefs = [
            [
                "man speaks",
                "man is speaking",
                "a man speaks",
                "man talks",
                "someone is talking",
            ],
            ["a bird is chirping"] * 5,
            ["heavy rain noise"] * 5,
        ]
        return cands, mrefs


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union

import torch

from torch import nn, Tensor
from torchmetrics.text.bert import _DEFAULT_MODEL

from aac_metrics.classes.base import AACMetric
from aac_metrics.functional.bert_score_mrefs import (
    bert_score_mrefs,
    _load_model_and_tokenizer,
)


class BERTScoreMRefs(AACMetric):
    """BERTScore metric which supports multiple references.

    The implementation is based on the bert_score implementation of torchmetrics.

    - Paper: https://arxiv.org/pdf/1904.09675.pdf

    For more information, see :func:`~aac_metrics.functional.bert_score.bert_score_mrefs`.
    """

    full_state_update = False
    higher_is_better = True
    is_differentiable = False

    min_value = 0.0
    max_value = 1.0

    def __init__(
        self,
        return_all_scores: bool = True,
        model: Union[str, nn.Module] = _DEFAULT_MODEL,
        device: Union[str, torch.device, None] = "auto",
        batch_size: int = 32,
        num_threads: int = 0,
        max_length: int = 64,
        reset_state: bool = True,
        idf: bool = False,
        reduction: str = "max",
        filter_nan: bool = True,
        verbose: int = 0,
    ) -> None:
        model, tokenizer = _load_model_and_tokenizer(
            model=model,
            tokenizer=None,
            device=device,
            reset_state=reset_state,
            verbose=verbose,
        )

        super().__init__()
        self._return_all_scores = return_all_scores
        self._model = model
        self._tokenizer = tokenizer
        self._device = device
        self._batch_size = batch_size
        self._num_threads = num_threads
        self._max_length = max_length
        self._reset_state = reset_state
        self._idf = idf
        self._reduction = reduction
        self._filter_nan = filter_nan
        self._verbose = verbose

        self._candidates = []
        self._mult_references = []

    def compute(self) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
        return bert_score_mrefs(
            candidates=self._candidates,
            mult_references=self._mult_references,
            return_all_scores=self._return_all_scores,
            model=self._model,
            tokenizer=self._tokenizer,
            device=self._device,
            batch_size=self._batch_size,
            num_threads=self._num_threads,
            max_length=self._max_length,
            reset_state=self._reset_state,
            idf=self._idf,
            reduction=self._reduction,
            filter_nan=self._filter_nan,
            verbose=self._verbose,
        )

    def extra_repr(self) -> str:
        if isinstance(self._model, str):
            model_name = self._model
        else:
            model_name = self._model.__class__.__name__

        hparams = {"model": model_name, "idf": self._idf}
        repr_ = ", ".join(f"{k}={v}" for k, v in hparams.items())
        return repr_

    def get_output_names(self) -> tuple[str, ...]:
        return (
            "bert_score.precision",
            "bert_score.recalll",
            "bert_score.f1",
        )

    def reset(self) -> None:
        self._candidates = []
        self._mult_references = []
        return super().reset()

    def update(
        self,
        candidates: list[str],
        mult_references: list[list[str]],
    ) -> None:
        self._candidates += candidates
        self._mult_references += mult_references



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
import shutil

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Union
from zipfile import ZipFile

from torch.hub import download_url_to_file

import aac_metrics

from aac_metrics.classes.bert_score_mrefs import BERTScoreMRefs
from aac_metrics.classes.fense import FENSE
from aac_metrics.functional.meteor import DNAME_METEOR_CACHE
from aac_metrics.functional.spice import (
    DNAME_SPICE_CACHE,
    DNAME_SPICE_LOCAL_CACHE,
    FNAME_SPICE_JAR,
    check_spice_install,
)
from aac_metrics.utils.cmdline import _str_to_bool, _setup_logging
from aac_metrics.utils.globals import (
    _get_cache_path,
    _get_tmp_path,
    get_default_cache_path,
    get_default_tmp_path,
)
from aac_metrics.utils.tokenization import FNAME_STANFORD_CORENLP_3_4_1_JAR


pylog = logging.getLogger(__name__)


DATA_URLS = {
    "meteor": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/meteor/meteor-1.5.jar",
        "fname": "meteor-1.5.jar",
    },
    "meteor_data": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/meteor/data/paraphrase-en.gz",
        "fname": osp.join("data", "paraphrase-en.gz"),
    },
    "meteor_data_fr": {
        "url": "https://github.com/cmu-mtlab/meteor/raw/master/data/paraphrase-fr.gz",
        "fname": osp.join("data", "paraphrase-fr.gz"),
    },
    "meteor_data_de": {
        "url": "https://github.com/cmu-mtlab/meteor/raw/master/data/paraphrase-de.gz",
        "fname": osp.join("data", "paraphrase-de.gz"),
    },
    "meteor_data_es": {
        "url": "https://github.com/cmu-mtlab/meteor/raw/master/data/paraphrase-es.gz",
        "fname": osp.join("data", "paraphrase-es.gz"),
    },
    "meteor_data_cz": {
        "url": "https://github.com/cmu-mtlab/meteor/raw/master/data/paraphrase-cz.gz",
        "fname": osp.join("data", "paraphrase-cz.gz"),
    },
    "spice": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/spice/spice-1.0.jar",
        "fname": "spice-1.0.jar",
    },
    "stanford_nlp": {
        "url": "https://github.com/tylin/coco-caption/raw/master/pycocoevalcap/tokenizer/stanford-corenlp-3.4.1.jar",
        "fname": "stanford-corenlp-3.4.1.jar",
    },
    "spice_zip": {
        "url": "https://panderson.me/images/SPICE-1.0.zip",
        "fname": "SPICE-1.0.zip",
    },
    "spice_corenlp_zip": {
        "url": "http://nlp.stanford.edu/software/stanford-corenlp-full-2015-12-09.zip",
        "fname": osp.join("SPICE-1.0", "stanford-corenlp-full-2015-12-09.zip"),
    },
}


def download_metrics(
    cache_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    clean_archives: bool = True,
    ptb_tokenizer: bool = True,
    meteor: bool = True,
    spice: bool = True,
    fense: bool = True,
    bert_score: bool = True,
    verbose: int = 0,
) -> None:
    """Download the code needed for SPICE, METEOR, PTB Tokenizer and FENSE.

    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param tmp_path: The path to a temporary directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param clean_archives: If True, remove all archives files. defaults to True.
    :param ptb_tokenizer: If True, downloads the PTBTokenizer code in cache directory. defaults to True.
    :param meteor: If True, downloads the METEOR code in cache directory. defaults to True.
    :param spice: If True, downloads the SPICE code in cache directory. defaults to True.
    :param fense: If True, downloads the FENSE models. defaults to True.
    :param bert_score: If True, downloads the BERTScore model. defaults to True.
    :param verbose: The verbose level. defaults to 0.
    """
    if verbose >= 1:
        pylog.info("aac-metrics download started.")

    cache_path = _get_cache_path(cache_path)
    tmp_path = _get_tmp_path(tmp_path)

    os.makedirs(cache_path, exist_ok=True)
    os.makedirs(tmp_path, exist_ok=True)

    if verbose >= 2:
        pylog.debug("AAC setup:")
        pylog.debug(f"  Cache directory: {cache_path}")
        pylog.debug(f"  Temp directory: {tmp_path}")

    if ptb_tokenizer:
        _download_ptb_tokenizer(cache_path, verbose)

    if meteor:
        _download_meteor(cache_path, verbose)

    if spice:
        _download_spice(cache_path, clean_archives, verbose)

    if fense:
        _download_fense(verbose)

    if bert_score:
        _download_bert_score(verbose)

    if verbose >= 1:
        pylog.info("aac-metrics download finished.")


def _download_ptb_tokenizer(
    cache_path: str,
    verbose: int = 0,
) -> None:
    # Download JAR file for tokenization
    stanford_nlp_dpath = osp.join(
        cache_path, osp.dirname(FNAME_STANFORD_CORENLP_3_4_1_JAR)
    )
    os.makedirs(stanford_nlp_dpath, exist_ok=True)

    name = "stanford_nlp"
    info = DATA_URLS[name]
    url = info["url"]
    fname = info["fname"]
    fpath = osp.join(stanford_nlp_dpath, fname)

    if not osp.isfile(fpath):
        if verbose >= 1:
            pylog.info(
                f"Downloading JAR source for '{name}' in directory {stanford_nlp_dpath}."
            )
        download_url_to_file(url, fpath, progress=verbose >= 1)
    else:
        if verbose >= 1:
            pylog.info(f"Stanford model file '{name}' is already downloaded.")


def _download_meteor(
    cache_path: str,
    verbose: int = 0,
) -> None:
    # Download JAR files for METEOR metric
    meteor_dpath = osp.join(cache_path, DNAME_METEOR_CACHE)
    os.makedirs(meteor_dpath, exist_ok=True)

    meteors_names = [name for name in DATA_URLS.keys() if name.startswith("meteor")]

    for name in meteors_names:
        info = DATA_URLS[name]
        url = info["url"]
        fname = info["fname"]
        subdir = osp.dirname(fname)
        fpath = osp.join(meteor_dpath, fname)

        if osp.isfile(fpath):
            if verbose >= 1:
                pylog.info(f"Meteor file '{name}' is already downloaded.")
            continue

        if verbose >= 1:
            pylog.info(f"Downloading source for '{fname}' in directory {meteor_dpath}.")
        if subdir not in ("", "."):
            os.makedirs(osp.dirname(fpath), exist_ok=True)

        download_url_to_file(
            url,
            fpath,
            progress=verbose >= 1,
        )


def _download_spice(
    cache_path: str,
    clean_archives: bool = True,
    verbose: int = 0,
) -> None:
    """Download SPICE java code.

    Target SPICE directory tree:

    spice
    ├── cache
    ├── lib
    │   ├── ejml-0.23.jar
    │   ├── fst-2.47.jar
    │   ├── guava-19.0.jar
    │   ├── hamcrest-core-1.3.jar
    │   ├── jackson-core-2.5.3.jar
    │   ├── javassist-3.19.0-GA.jar
    │   ├── json-simple-1.1.1.jar
    │   ├── junit-4.12.jar
    │   ├── lmdbjni-0.4.6.jar
    │   ├── lmdbjni-linux64-0.4.6.jar
    │   ├── lmdbjni-osx64-0.4.6.jar
    │   ├── lmdbjni-win64-0.4.6.jar
    │   ├── Meteor-1.5.jar
    │   ├── objenesis-2.4.jar
    │   ├── SceneGraphParser-1.0.jar
    │   ├── slf4j-api-1.7.12.jar
    │   ├── slf4j-simple-1.7.21.jar
    │   ├── stanford-corenlp-3.6.0.jar
    │   └── stanford-corenlp-3.6.0-models.jar
    └── spice-1.0.jar
    """
    try:
        check_spice_install(cache_path)
        return None
    except (FileNotFoundError, NotADirectoryError, PermissionError):
        pass

    # Download JAR files for SPICE metric
    spice_cache_dpath = osp.join(cache_path, DNAME_SPICE_CACHE)
    spice_jar_dpath = osp.join(cache_path, osp.dirname(FNAME_SPICE_JAR))
    spice_local_cache_path = osp.join(cache_path, DNAME_SPICE_LOCAL_CACHE)

    os.makedirs(spice_jar_dpath, exist_ok=True)
    os.makedirs(spice_local_cache_path, exist_ok=True)

    for name in ("spice_zip", "spice_corenlp_zip"):
        url = DATA_URLS[name]["url"]
        fname = DATA_URLS[name]["fname"]
        fpath = osp.join(spice_cache_dpath, fname)

        if osp.isfile(fpath):
            if verbose >= 1:
                pylog.info(f"File '{fpath}' is already downloaded for SPICE.")
        else:
            if verbose >= 1:
                pylog.info(f"Downloading file '{fpath}' for SPICE...")

            dpath = osp.dirname(fpath)
            os.makedirs(dpath, exist_ok=True)
            download_url_to_file(url, fpath, progress=verbose > 0)

        if fname.endswith(".zip"):
            if verbose >= 1:
                pylog.info(f"Extracting {fname} to {spice_cache_dpath}...")

            with ZipFile(fpath, "r") as file:
                file.extractall(spice_cache_dpath)

    spice_lib_dpath = osp.join(spice_cache_dpath, "lib")
    spice_unzip_dpath = osp.join(spice_cache_dpath, "SPICE-1.0")
    corenlp_dpath = osp.join(spice_cache_dpath, "stanford-corenlp-full-2015-12-09")

    # Note: order matter here
    to_move = [
        ("f", osp.join(spice_unzip_dpath, "spice-1.0.jar"), spice_cache_dpath),
        ("f", osp.join(corenlp_dpath, "stanford-corenlp-3.6.0.jar"), spice_lib_dpath),
        (
            "f",
            osp.join(corenlp_dpath, "stanford-corenlp-3.6.0-models.jar"),
            spice_lib_dpath,
        ),
    ]
    for name in os.listdir(osp.join(spice_unzip_dpath, "lib")):
        if not name.endswith(".jar"):
            continue
        fpath = osp.join(spice_unzip_dpath, "lib", name)
        to_move.append(("f", fpath, spice_lib_dpath))

    os.makedirs(spice_lib_dpath, exist_ok=True)

    for i, (_src_type, src_path, parent_tgt_dpath) in enumerate(to_move):
        tgt_path = osp.join(parent_tgt_dpath, osp.basename(src_path))

        if osp.exists(tgt_path):
            if verbose >= 1:
                pylog.info(
                    f"Target '{tgt_path}' already exists. ({i+1}/{len(to_move)})"
                )
        else:
            if verbose >= 1:
                pylog.info(
                    f"Moving '{src_path}' to '{parent_tgt_dpath}'... ({i+1}/{len(to_move)})"
                )
            shutil.move(src_path, parent_tgt_dpath)

    shutil.rmtree(corenlp_dpath)
    if clean_archives:
        spice_zip_fname = DATA_URLS["spice_zip"]["fname"]
        spice_zip_fpath = osp.join(spice_cache_dpath, spice_zip_fname)

        os.remove(spice_zip_fpath)
        shutil.rmtree(spice_unzip_dpath)


def _download_fense(
    verbose: int = 0,
) -> None:
    # Download models files for FENSE metric
    if verbose >= 1:
        pylog.info("Downloading SBERT and BERT error detector for FENSE metric...")
    _ = FENSE(device="cpu")


def _download_bert_score(
    verbose: int = 0,
) -> None:
    # Download models files for BERTScore metric
    if verbose >= 1:
        pylog.info("Downloading BERT model for BERTScore metric...")
    _ = BERTScoreMRefs(device="cpu")


def _get_main_download_args() -> Namespace:
    parser = ArgumentParser(
        description="Download models and external code to evaluate captions."
    )

    parser.add_argument(
        "--cache_path",
        type=str,
        default=get_default_cache_path(),
        help="Cache directory path.",
    )
    parser.add_argument(
        "--tmp_path",
        type=str,
        default=get_default_tmp_path(),
        help="Temporary directory path.",
    )
    parser.add_argument(
        "--clean_archives",
        type=_str_to_bool,
        default=True,
        help="If True, remove all archives files. defaults to True.",
    )
    parser.add_argument(
        "--ptb_tokenizer",
        type=_str_to_bool,
        default=True,
        help="Download PTBTokenizer Java source code.",
    )
    parser.add_argument(
        "--meteor",
        type=_str_to_bool,
        default=True,
        help="Download METEOR Java source code.",
    )
    parser.add_argument(
        "--spice",
        type=_str_to_bool,
        default=True,
        help="Download SPICE Java source code.",
    )
    parser.add_argument(
        "--fense",
        type=_str_to_bool,
        default=True,
        help="Download FENSE models.",
    )
    parser.add_argument("--verbose", type=int, default=1, help="Verbose level.")

    args = parser.parse_args()
    return args


def _main_download() -> None:
    args = _get_main_download_args()
    _setup_logging(aac_metrics.__package__, args.verbose)

    download_metrics(
        cache_path=args.cache_path,
        tmp_path=args.tmp_path,
        clean_archives=args.clean_archives,
        ptb_tokenizer=args.ptb_tokenizer,
        meteor=args.meteor,
        spice=args.spice,
        fense=args.fense,
        verbose=args.verbose,
    )


if __name__ == "__main__":
    _main_download()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import defaultdict, Counter
from typing import Any, Callable, Mapping, Union

import numpy as np
import torch

from torch import Tensor

from aac_metrics.utils.checks import check_metric_inputs


def cider_d(
    candidates: list[str],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    n: int = 4,
    sigma: float = 6.0,
    tokenizer: Callable[[str], list[str]] = str.split,
    return_tfidf: bool = False,
    scale: float = 10.0,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Any]]]:
    """Consensus-based Image Description Evaluation function.

    - Paper: https://arxiv.org/pdf/1411.5726.pdf

    .. warning::
        This metric requires at least 2 candidates with 2 sets of references, otherwise it will raises a ValueError.

    :param candidates: The list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param n: Maximal number of n-grams taken into account. defaults to 4.
    :param sigma: Standard deviation parameter used for gaussian penalty. defaults to 6.0.
    :param tokenizer: The fast tokenizer used to split sentences into words. defaults to str.split.
    :param return_tfidf: If True, returns the list of dictionaries containing the tf-idf scores of n-grams in the sents_score output.
        defaults to False.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    cooked_cands, cooked_mrefs = _cider_d_update(
        candidates,
        mult_references,
        n,
        tokenizer,
        [],
        [],
    )
    return _cider_d_compute(
        cooked_cands,
        cooked_mrefs,
        return_all_scores,
        n,
        sigma,
        return_tfidf,
        scale,
    )


def _cider_d_update(
    candidates: list[str],
    mult_references: list[list[str]],
    n: int,
    tokenizer: Callable[[str], list[str]],
    prev_cooked_cands: list[Counter],
    prev_cooked_mrefs: list[list[Counter]],
) -> tuple[list, list]:
    check_metric_inputs(candidates, mult_references)
    new_cooked_mrefs = [
        [__cook_sentence(ref, n, tokenizer) for ref in refs] for refs in mult_references
    ]
    new_cooked_cands = [__cook_sentence(cand, n, tokenizer) for cand in candidates]
    prev_cooked_cands += new_cooked_cands
    prev_cooked_mrefs += new_cooked_mrefs
    return prev_cooked_cands, prev_cooked_mrefs


def _cider_d_compute(
    cooked_cands: list[Counter],
    cooked_mrefs: list[list[Counter]],
    return_all_scores: bool,
    n: int,
    sigma: float,
    return_tfidf: bool,
    scale: float,
) -> Union[Tensor, tuple[dict[str, Tensor], dict[str, Any]]]:
    if len(cooked_cands) <= 1:
        raise ValueError(
            f"CIDEr-D metric does not support less than 2 candidates with 2 references. (found {len(cooked_cands)} candidates, but expected > 1)"
        )
    # compute idf
    doc_frequencies = __compute_doc_freq(cooked_mrefs)
    # sanity check: assert to check document frequency
    assert len(cooked_cands) >= max(doc_frequencies.values()), "Sanity check failed."

    # compute log reference length
    log_n_refs = np.log(float(len(cooked_mrefs)))
    # compute cider score
    cider_d_scores, tfidf_lst = __compute_cider(
        cooked_cands,
        cooked_mrefs,
        doc_frequencies,
        log_n_refs,
        n,
        sigma,
        scale=scale,
    )
    cider_d_score = cider_d_scores.mean()

    cider_d_scores = torch.from_numpy(cider_d_scores)
    cider_d_score = torch.as_tensor(cider_d_score, dtype=torch.float64)

    if return_all_scores:
        cider_d_outs_corpus = {
            "cider_d": cider_d_score,
        }
        cider_d_outs_sents = {
            "cider_d": cider_d_scores,
        }
        if return_tfidf:
            cider_d_outs_sents["tfidf_lst"] = tfidf_lst  # type: ignore
        cider_d_outs = cider_d_outs_corpus, cider_d_outs_sents

        return cider_d_outs
    else:
        return cider_d_score


def __cook_sentence(
    sentence: str,
    n: int,
    tokenizer: Callable[[str], list[str]],
) -> Counter[tuple[str, ...]]:
    """
    Takes a string as input and returns an object that can be given to
    either cook_refs or cook_test. This is optional: cook_refs and cook_test
    can take string arguments as well.
    :param s: string : sentence to be converted into ngrams
    :param n: int    : number of ngrams for which representation is calculated
    :return: term frequency vector for occuring ngrams
    """
    words = tokenizer(sentence)
    counter = Counter()
    for k in range(1, n + 1):
        for i in range(len(words) - k + 1):
            ngram = tuple(words[i : i + k])
            counter[ngram] += 1
    return counter


def __compute_doc_freq(cooked_mrefs: list[list[Counter]]) -> Counter[tuple[str, ...]]:
    """
    Compute term frequency for reference data.
    This will be used to compute idf (inverse document frequency later)
    The term frequency is stored in the object
    :return: None
    """
    doc_frequencies = Counter()
    for refs in cooked_mrefs:
        all_refs_ngrams = set(ngram for ref in refs for ngram in ref.keys())
        for ngram in all_refs_ngrams:
            doc_frequencies[ngram] += 1

    return doc_frequencies


def __counter_to_vec(
    counters: dict[tuple, int],
    log_n_refs: float,
    n: int,
    doc_frequencies: Union[Mapping[tuple, int], Callable[[tuple], int]],
) -> tuple[list[defaultdict], np.ndarray, int]:
    """
    Function maps counts of ngram to vector of tfidf weights.
    The function returns vec, an array of dictionary that store mapping of n-gram and tf-idf weights.
    The n-th entry of array denotes length of n-grams.
    :param cnts:
    :return: tf-idf of n-grams (array of n dict[tuple, float]), norm (array of n floats) (norms for n-grams), length (int) (number of distinct n-grams from 1 to n)
    """
    vec = [defaultdict(float) for _ in range(n)]
    length = 0
    norm = np.zeros((n,))

    for ngram, term_freq in counters.items():
        if isinstance(doc_frequencies, Mapping):
            count = doc_frequencies[ngram]
        else:
            count = doc_frequencies(ngram)

        # give ngram count 1 if it doesn't appear in reference corpus
        log_df = np.log(max(1.0, count))

        # ngram index
        cur_n = len(ngram) - 1

        # tf (term_freq) * idf (precomputed idf) for n-grams
        vec[cur_n][ngram] = float(term_freq) * (log_n_refs - log_df)

        # compute norm for the vector.  the norm will be used for computing similarity
        norm[cur_n] += pow(vec[cur_n][ngram], 2)

        if cur_n == 1:
            length += term_freq

    norm = np.sqrt(norm)
    return vec, norm, length


def __similarity(
    cand_vec: list[defaultdict],
    ref_vec: list[defaultdict],
    cand_norm: np.ndarray,
    ref_norm: np.ndarray,
    cand_len: int,
    ref_len: int,
    n: int,
    sigma: float,
) -> np.ndarray:
    """
    Compute the cosine similarity of two vectors.

    :param cand_vec: (n, nb_ngrams_of_len_n), contains the TFIDF scores for one candidate
    :param ref_vec: (n, nb_ngrams_of_len_n), contains the TFIDF scores for one reference
    :param cand_norm: (n,), norms of the candidate n-grams vectors
    :param ref_norm: (n,), norms of the reference n-grams vectors
    :param cand_len: Size of the candidate
    :returns: N-grams similarities as array of shape (n,)
    """
    delta = float(cand_len - ref_len)
    # measure consine similarity
    similarities = np.zeros((n,))

    for ni in range(n):
        # ngram
        for ngram, count in cand_vec[ni].items():
            # vrama91 : added clipping
            similarities[ni] += min(count, ref_vec[ni][ngram]) * ref_vec[ni][ngram]

        if (cand_norm[ni] != 0) and (ref_norm[ni] != 0):
            similarities[ni] /= cand_norm[ni] * ref_norm[ni]

        # vrama91: added a length based gaussian penalty
        similarities[ni] *= np.e ** (-(delta**2) / (2 * sigma**2))

    return similarities


def __compute_cider(
    cooked_cands: list[Counter],
    cooked_mrefs: list[list[Counter]],
    doc_frequencies: Union[Counter[tuple], Callable[[tuple], int]],
    log_n_refs: float,
    n: int,
    sigma: float,
    scale: float,
) -> tuple[np.ndarray, list[tuple[list, list]]]:
    scores = np.empty((len(cooked_cands),))
    tfidf_lst = []

    for i, (cand, refs) in enumerate(zip(cooked_cands, cooked_mrefs)):
        # compute vector for test captions
        vec, norm, length = __counter_to_vec(cand, log_n_refs, n, doc_frequencies)
        # compute vector for ref captions
        ngrams_scores = np.zeros((len(refs), n))
        vec_refs = []
        for j, ref in enumerate(refs):
            vec_ref, norm_ref, length_ref = __counter_to_vec(
                ref, log_n_refs, n, doc_frequencies
            )
            vec_refs.append(vec_ref)
            ngrams_scores[j] = __similarity(
                vec, vec_ref, norm, norm_ref, length, length_ref, n, sigma
            )

        # Use this weird mean calculation instead of ".mean()" because it can give slight differences compared to the original implementation
        score_avg = ngrams_scores.sum(axis=0).mean() / len(refs)
        scores[i] = score_avg
        tfidf_lst.append((vec, vec_refs))

    scores = scores * scale
    return scores, tfidf_lst



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Callable, Iterable, Optional, Union

import torch

from torch import Tensor

from aac_metrics.functional.mult_cands import mult_cands_metric
from aac_metrics.functional.spider import spider


def spider_max(
    mult_candidates: list[list[str]],
    mult_references: list[list[str]],
    return_all_scores: bool = True,
    return_all_cands_scores: bool = False,
    # CIDEr args
    n: int = 4,
    sigma: float = 6.0,
    tokenizer: Callable[[str], list[str]] = str.split,
    return_tfidf: bool = False,
    # SPICE args
    cache_path: Union[str, Path, None] = None,
    java_path: Union[str, Path, None] = None,
    tmp_path: Union[str, Path, None] = None,
    n_threads: Optional[int] = None,
    java_max_memory: str = "8G",
    timeout: Union[None, int, Iterable[int]] = None,
    verbose: int = 0,
) -> Union[tuple[dict[str, Tensor], dict[str, Tensor]], Tensor]:
    """SPIDEr-max function.

    Compute the maximal SPIDEr score accross multiple candidates.

    - Paper: https://dcase.community/documents/workshop2022/proceedings/DCASE2022Workshop_Labbe_46.pdf

    .. warning::
        This metric requires at least 2 candidates with 2 sets of references, otherwise it will raises a ValueError.

    :param mult_candidates: The list of list of sentences to evaluate.
    :param mult_references: The list of list of sentences used as target.
    :param return_all_scores: If True, returns a tuple containing the globals and locals scores.
        Otherwise returns a scalar tensor containing the main global score.
        defaults to True.
    :param return_all_cands_scores: If True, returns all multiple candidates scores in sents_scores outputs as tensor of shape (n_audio, n_cands_per_audio).
        defaults to False.
    :param n: Maximal number of n-grams taken into account. defaults to 4.
    :param sigma: Standard deviation parameter used for gaussian penalty. defaults to 6.0.
    :param tokenizer: The fast tokenizer used to split sentences into words. defaults to str.split.
    :param return_tfidf: If True, returns the list of dictionaries containing the tf-idf scores of n-grams in the sents_score output.
        defaults to False.
    :param cache_path: The path to the external code directory. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_cache_path`.
    :param java_path: The path to the java executable. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_java_path`.
    :param tmp_path: Temporary directory path. defaults to the value returned by :func:`~aac_metrics.utils.paths.get_default_tmp_path`.
    :param java_max_memory: The maximal java memory used. defaults to "8G".
    :param n_threads: Number of threads used to compute SPICE.
        None value will use the default value of the java program.
        defaults to None.
    :param timeout: The number of seconds before killing the java subprogram.
        If a list is given, it will restart the program if the i-th timeout is reached.
        If None, no timeout will be used.
        defaults to None.
    :param verbose: The verbose level. defaults to 0.
    :returns: A tuple of globals and locals scores or a scalar tensor with the main global score.
    """
    return mult_cands_metric(
        metric=spider,
        metric_out_name="spider",
        mult_candidates=mult_candidates,
        mult_references=mult_references,
        return_all_scores=return_all_scores,
        return_all_cands_scores=return_all_cands_scores,
        selection="max",
        reduction=torch.mean,
        # CIDEr args
        n=n,
        sigma=sigma,
        tokenizer=tokenizer,
        return_tfidf=return_tfidf,
        # SPICE args
        cache_path=cache_path,
        java_path=java_path,
        tmp_path=tmp_path,
        n_threads=n_threads,
        java_max_memory=java_max_memory,
        timeout=timeout,
        verbose=verbose,
    )
